@@ -15,6 +15,8 @@ foreach gurkmeja [array names auto_index] {
 if {[info exists gurkmeja]} {
     unset gurkmeja
 }
+# Remove this that might leak in from package install
+catch {rename tkpath::load_package {}}
 
 
 # First get some data about the system
@@ -67,6 +69,18 @@ if 1 { # Not working yet?
 proc getSubCmds {args} {
     catch {uplevel 1 $args} err
 
+    # Some commands show their available options in this message
+    if {[regexp "wrong \# args:" $err]} {
+        #puts "$args"
+        #puts "  $err"
+        set result {}
+        foreach {_ x} [regexp -all -inline {\?(-\w+)} $err] {
+            lappend result $x
+        }
+        #puts "  $result"
+        return $result
+    }        
+
     lappend regexps {option .* must be (.*)$}
     lappend regexps {option .* should be one of (.*)$}
     lappend regexps {bad .* must be (.*)$}
@@ -157,46 +171,59 @@ proc buildDb {ch} {
     # If a token is an integer, just check the number of arguments against it.
     # r min ?max?  Specifies a range for number of arguments
 
-    # x  Any
-    # o  Option, i.e anything starting with -
-    # p  Option+Any (p as in option Pair)
-    # s  Subcommand
-    # d* Definitions. That arg defines a new command, as detailed below
-    # di Define inheritance
-    # dc Define with copy, if followed by =cmd, it copies syntax from cmd.
-    # do Define object, if followed by =cmd, it copies syntax from cmd.
-    # dk Define constructor (args+body)
-    # dd Define destructor (just body)
-    # dp Define procedure (name+args+body)
-    # dm Define method (name+args+body)
+    # The following token names are recognised. A token name is alphabetic
+    # and may be mixed case. A token may cover and consume multiple arguments.
+    # x   Any
+    # o   Option, i.e anything starting with -.
+    #     An option may consume a second arg if the option database says so.
+    # p   Option+Any i.e. (o x) (p as in option Pair)
+    # s   Subcommand
+    # d*  Definitions. That arg defines a new command, as detailed below
+    # di  Define inheritance
+    # dc  Define with copy, if followed by =cmd, it copies syntax from cmd.
+    # do  Define object, if followed by =cmd, it copies syntax from cmd.
+    # dk  Define constructor (args+body)
+    # dd  Define destructor (just body)
+    # dp  Define procedure (name+args+body)
+    # dm  Define method (name+args+body)
     # dmp Define metod/procedure (name+args+body)
     # div Define implicit variable
-    # e  Expression
-    # E  Expression that should be in braces
-    # c  Code, checked in surrounding context
-    #    If an integer is added to it, that number of argumenst are added to the
-    #    code to emulate a command prefix. (cg has this too)
-    # cg Code, checked in global context
-    # cn Code, checked in virtual namespace
-    # cl Code, checked in its own local context
-    # cv Code, checked in its own local context, preceded by variable list
-    # n, v and l all marks variable names. Those arguments will not be
-    #   checked against known variables to detect missing $.
-    # n  The variable does not have to exist, and is set by the command.
-    # v  The variable must exist. It is not marked as set.
-    # l  Does not have to exist. It will be marked as known, but not set.
-    # nl Like n, but a list of variable names
+    # e   Expression
+    # E   Expression that should be in braces
+    # re  regexp
+    # c   Code, checked in surrounding context
+    #     If an integer is added to it, that number of arguments are added to
+    #     the code to emulate a command prefix. (cg has this too)
+    # cg  Code, checked in global context
+    # cn  Code, checked in virtual namespace
+    # cl  Code, checked in its own local context
+    # cv  Code, checked in its own local context, preceded by variable list
+    #     n, v and l all marks variable names. Those arguments will not be
+    #     checked against known variables to detect missing $.
+    # n   The variable does not have to exist, and is set by the command.
+    # v   The variable must exist. It is not marked as set.
+    # l   Does not have to exist. It will be marked as known, but not set.
+    # nl  Like n, but a list of variable names
 
     # Modifiers that apply to some of the above
     # ? Zero or One
-    # * Zero or more
+    # * Zero or more  (not supported by all tokens)
     # . One or nothing at all
 
-    # * after x swallows all the rest and must be last
-    # s may only have .
-    # e and c may not have any modifier
+    # If a token is followed by a parenthesis it denotes a type.
+    # The local variable in the proc is marked with this type.
+    # Any modifier goes after the parens.
+    # E.g.   x(varName)   x(script)?
+    
+    # If a token is followed by a number it is token dependent.
+    # Any modifier goes after the number.
+    # E.g.   c2   cg4?
+
+    # If a token is followed by =, it is a token dependent modifier.
+    # E.g   do=_stdclass_oo
 
     # If a syntax for a subcommand is defined, it is used to check the rest
+    # If a syntax for an option is defined, it is used to check the next arg.
 
 
     # Syntax for Tcl core commands
@@ -321,8 +348,8 @@ proc buildDb {ch} {
     set syntax(puts)            "1: x : o? x x?"
     set syntax(pwd)              0
     set syntax(read)            "r 1 2"
-    set syntax(regexp)          "o* x x n*"
-    set syntax(regsub)          "o* x x x n?"
+    set syntax(regexp)          "o* re x n*"
+    set syntax(regsub)          "o* re x x n?"
     set syntax(rename)           2   ;# Maybe treat rename specially?
     set syntax(return)          "p* x?"
     set syntax(scan)            "x x n*"
@@ -393,15 +420,26 @@ proc buildDb {ch} {
         set syntax(catch)        "c n? n?" ;# FIXA make a test for this
         set syntax(dict)          "s x x*"
         set syntax(dict\ append)  "n x x*"
-        set syntax(dict\ incr)    "n x x*"
-        set syntax(dict\ filter)  "x x x"
-        set syntax(dict\ lappend) "n x x*"
-        set syntax(dict\ set)     "n x x*"
-        set syntax(dict\ unset)   "n x x*"
-        # FIXA: handle this style:
-        set syntax(dict\ update)  "l x x x* c"
-        set syntax(dict\ with)    "l x* c"
+        set syntax(dict\ create)  "x&x*"
+        set syntax(dict\ exists)  "x x x*"
         set syntax(dict\ for)     "nl x c"
+        set syntax(dict\ incr)    "n x x*"
+        set syntax(dict\ filter)  "x s x*"
+        set syntax(dict\ filter\ key)    "x*"
+        set syntax(dict\ filter\ script) "nl c"
+        set syntax(dict\ filter\ value)  "x*"
+        set syntax(dict\ info)    "x"
+        set syntax(dict\ keys)    "x x?"
+        set syntax(dict\ lappend) "n x x*"
+        set syntax(dict\ map)     "nl x c"
+        set syntax(dict\ remove)  "x x*"
+        set syntax(dict\ replace) "x x*"
+        set syntax(dict\ set)     "n x x x*"
+        set syntax(dict\ size)    "x"
+        set syntax(dict\ unset)   "n x x*"
+        set syntax(dict\ values)  "x x?"
+        set syntax(dict\ update)  "l x n x&n* c"
+        set syntax(dict\ with)    "l x* c"
         
         # Initialising incr
         set syntax(incr)       "n x?"
@@ -571,7 +609,48 @@ proc buildDb {ch} {
         # New options
         set option(lsort\ -stride) 1
     }
-        
+
+    # Things added in 8.7
+    if {[info commands lpop] ne ""} {
+        set syntax(lpop) "v x*"
+        set syntax(timerate) "o* c x? x?"
+        set option(timerate\ -overhead) 1
+
+        set syntax(zipfs) "s x*"
+        set syntax(zipfs\ canonical) "r 1 3"
+        set syntax(zipfs\ exists)  1
+        set syntax(zipfs\ find)    1
+        set syntax(zipfs\ info)    1
+        set syntax(zipfs\ list) "o* x?"
+        set option(zipfs\ list) "-glob -regexp"
+        set syntax(zipfs\ lmkimg)  "r 2 4"
+        set syntax(zipfs\ lmkzip)  "r 2 3"
+        set syntax(zipfs\ mkimg)   "r 2 5"
+        set syntax(zipfs\ mkkey)   1
+        set syntax(zipfs\ mkzip)   "r 2 4"
+        set syntax(zipfs\ mount)   "r 0 3"
+        set syntax(zipfs\ root)    0
+        set syntax(zipfs\ unmount) 1
+
+        set syntax(tcl::process) "s x*"
+        set syntax(tcl::process\ autoperge) "x?"
+        set syntax(tcl::process\ list)      "0"
+        set syntax(tcl::process\ purge)     "x?"
+        set syntax(tcl::process\ status)    "o* x?"
+
+        set syntax(array\ default) "s l x?"
+        set syntax(array\ default\ exists) "l=array"
+        set syntax(array\ default\ get)    "v=array"
+        set syntax(array\ default\ set)    "v=array x"
+        set syntax(array\ default\ unset)  "v=array"
+
+        set syntax(array\ for) "nl v=array c"
+
+        set syntax(info\ cmdtype) "x"
+
+        set option(lsearch\ -stride) 1
+        set option(regsub\ -command) 1
+    }
 
     # Some special Tcl commands
     set syntax(dde)             "o? s x*"  ;# FIXA: is this correct?
@@ -656,7 +735,7 @@ proc buildDb {ch} {
             set syntax(ttk::style\ element\ options) x
             set syntax(ttk::style\ theme) "s x*"
             set syntax(ttk::style\ theme\ create) "x p*"
-            set syntax(ttk::style\ theme\ settings) 2
+            set syntax(ttk::style\ theme\ settings) "x c"
             set syntax(ttk::style\ theme\ names) 0
             set syntax(ttk::style\ theme\ use) x
             set syntax(ttk::themes) x?
@@ -685,12 +764,79 @@ proc buildDb {ch} {
             set option(_obj,$class\ configure) $option($class)
             set syntax(_obj,$class\ cget) "o"
             set option(_obj,$class\ cget) $option($class)
+            # Widget details
             switch $class {
                 listbox {
                     set syntax(_obj,$class\ selection) "s x x?"
                     set subCmd(_obj,$class\ selection) \
                             [getSubCmds .w selection gurkmeja 0]
                 }
+                text {
+                    set syntax(_obj,$class\ bbox) "1"
+                    set syntax(_obj,$class\ compare) "3"
+                    set syntax(_obj,$class\ debug) "x?"
+                    set syntax(_obj,$class\ delete) "r 1"
+                    set syntax(_obj,$class\ dlineinfo) "1"
+                    set syntax(_obj,$class\ index) "1"
+                    set syntax(_obj,$class\ insert) "r 2"
+                    set syntax(_obj,$class\ pendingsync) "0"
+                    set syntax(_obj,$class\ replace) "r 3"
+                    set syntax(_obj,$class\ see) "1"
+                    set syntax(_obj,$class\ sync) "r 1 3"
+
+                    set syntax(_obj,$class\ edit) "s x*"
+                    set subCmd(_obj,$class\ edit) \
+                            [getSubCmds .w edit gurkmeja]
+
+                    set syntax(_obj,$class\ image) "s x*"
+                    set subCmd(_obj,$class\ image) \
+                            [getSubCmds .w image gurkmeja]
+
+                    set syntax(_obj,$class\ mark) "s x*"
+                    set subCmd(_obj,$class\ mark) \
+                            [getSubCmds .w mark gurkmeja]
+
+                    set syntax(_obj,$class\ peer) "s x*"
+                    set subCmd(_obj,$class\ peer) \
+                            [getSubCmds .w peer gurkmeja]
+
+                    set syntax(_obj,$class\ scan) "s x x"
+                    set subCmd(_obj,$class\ scan) \
+                            [getSubCmds .w scan gurkmeja x x]
+
+                    set syntax(_obj,$class\ tag) "s x*"
+                    set subCmd(_obj,$class\ tag) \
+                            [getSubCmds .w tag gurkmeja]
+
+                    set syntax(_obj,$class\ window) "s x*"
+                    set subCmd(_obj,$class\ window) \
+                            [getSubCmds .w window gurkmeja]
+
+                    set syntax(_obj,$class\ xview) "0: 0 : s x*"
+                    set subCmd(_obj,$class\ xview) \
+                            [getSubCmds .w xview gurkmeja]
+
+                    # yview is a bit weird
+                    set syntax(_obj,$class\ yview) "x*"
+
+                    set syntax(_obj,$class\ count) "o* x x"
+                    set option(_obj,$class\ count) \
+                            [getSubCmds .w count -gurkmeja x x]
+
+                    set syntax(_obj,$class\ dump) "o* x x?"
+                    set option(_obj,$class\ dump) \
+                            [getSubCmds .w dump -gurkmeja x]
+                    set option(_obj,$class\ dump\ -command) x
+
+                    set syntax(_obj,$class\ get) "o* x x*"
+                    set option(_obj,$class\ get) \
+                            [getSubCmds .w get -gurkmeja x x]
+
+                    set syntax(_obj,$class\ search) "o* x x x?"
+                    set option(_obj,$class\ search) \
+                            [getSubCmds .w search -gurkmeja x x]
+                    set option(_obj,$class\ search\ -count) n
+                }
             }
         }
         set option(.\ configure)       $option(toplevel)
@@ -780,14 +926,6 @@ proc buildDb {ch} {
         }
     }
 
-    # A fix since puts still gives an unhelpful error
-    if {![info exists option(puts)] || [lsearch $option(puts) "-nonewline"] < 0} {
-        set option(puts) [list -nonewline]
-        # Also chan puts if present
-        if {[info exists syntax(chan\ puts)]} {
-            set option(chan\ puts) [list -nonewline]
-        }            
-    }
     # Send cannot return info in some versions
     if {![info exists option(send)]} {
         set option(send) "-- -async -displayof"
@@ -837,6 +975,36 @@ proc buildDb {ch} {
         }
     }
 
+    # Special handling of tcl::math*
+    foreach ns {mathop mathfunc} {
+        foreach fun [info commands tcl::${ns}::*] {
+            # Remove ::
+            set fun [string range $fun 2 end]
+            #if {[info exists syntax($fun)]} continue
+            # Try to figure it out
+            set minArg 0
+            set maxArg 1000
+            set e0 [catch {$fun}]
+            set e1 [catch {$fun 1}]
+            set e2 [catch {$fun 1 2}]
+            set e5 [catch {$fun 1 2 1 2 1}]
+            set spec ""
+            switch $e0$e1$e2$e5 {
+                0111 {set spec 0 }
+                1011 {set spec 1 }
+                1101 {set spec 2 }
+                0000 {set spec "r 0"}
+                1000 {set spec "r 1"}
+                default {
+                    puts "$fun $e0 $e1 $e2 $e5"
+                }
+            }
+            if {$spec ne ""} {
+                set syntax($fun) $spec
+            }
+        }
+    }
+
     # Output the data
     puts $ch [list lappend ::dbInfo $dbstring]
     puts $ch [list set ::dbTclVersion $::tcl_version]
