@@ -47,7 +47,7 @@ menu .m.f
 menu .m.s
 textSearch::searchMenu .m.s
 
-    if {![catch {set ch [open /usr/local/share/doc/textsearch/readme.txt r]}]} {
+    if {![catch {set ch [open [file join [file dirname [::tcl::pkgconfig get libdir,install]] share doc textsearch readme.txt] r]}]} {
         $w insert end [read $ch]
         close $ch
     }
